;
; This algorithm tries to implement an iterative blind deconvolution
; where for each iteration the decision if the target or psf is
; improved is based ot the effort versus the improvement (compared
; by using the convolution and the measured object).
; psf(i)  := psf at the iteration step i
; tar(i)  := target at iteration step i
; obj     := measurement of the target
; psf(i+1) = RL(obj, tar(i))  improvement of the psf
; tar(i+1) = RL(obj, psf(i))  improvement of the target
; obj(p) = tar(i) x psf(i+1)  old target convolved with new psf
; obj(t) = tar(i+1) x psf(i)  new target convolved with old psf
; The decision if the psf or the target will be improved can be based
; on different tests:
; 1. Simply compare the difference to the measured object
;   IF ERR(obj, obj(p)) < ERR(obj, obj(t))
;     THEN use the new psf
;     ELSE use the new target
;   ENDIF
; 2. Use the difference o the measured object related to the changes
; of the psf and target
;   IF ERR(obj, obj(p))/DIF(psf(i), psf(i+1)) < ...
;     THEN use the new psf
;     ELSE use the new target
;   ENDIF
FUNCTION ldrsDECONV_IBD::Init, algorithm
IF NOT self->ldrsDECONVOLVE::Init(algorithm) THEN RETURN, 0

RETURN, 1
END

FUNCTION ldrsDECONV_IBD::Prepare
PRINT, 'ldrsDECONV_IBD::Prepare'
IF NOT self->ldrsDECONVOLVE::Prepare() THEN RETURN, 0
PRINT, 'ldrsDECONV_IBD::Prepare', self.nx, self.ny, self.nPA
s = FLOAT(self.nx)*FLOAT(self.ny)
; create exactly one PSF array and put all position angles into that array
; (remember the scaling!)
self.psf = PTR_NEW(FLTARR(self.nx, self.ny))
self.psf1 = PTR_NEW(FLTARR(self.nx, self.ny))
nFiles = self.psfFT->GetNFiles()
FOR iFile=0,nFiles-1 DO BEGIN
    file = self.psfFT->GetFile(iFile)
    IF file->LoadData() EQ 0 THEN RETURN, 0
    data = file->GetData()
    FOR iZ=0, file->GetNZ()-1 DO BEGIN
        (*self.psf)[*,*] += (*data)[*,*,iZ]*(1.0/TOTAL((*data)[*,*,iZ], /DOUBLE))
    ENDFOR
ENDFOR
index = WHERE(*self.psf LT 0.0, count)
IF count NE 0 THEN (*self.psf)[index] = 0.0
(*self.psf)[*,*] *= 1.0/TOTAL(*self.psf, /DOUBLE) ; normalize it to integral 1.0
; calculate the average number of photons in the input raw files
; (excluding the background)
nFiles = self.objFT->GetNFiles()
self.nPhot = 0.0
self.bgrd = 0.0
bgrd = 0.0
count = 0.0
FOR iFile=0,nFiles-1 DO BEGIN
    file = self.objFT->GetFile(iFile)
    IF file->LoadData() EQ 0 THEN RETURN, 0
    data = file->GetData()
    FOR iZ=0, file->GetNZ()-1 DO BEGIN
        cfg = file->GetImageCFG(iZ)
        IF OBJ_VALID(cfg) THEN BEGIN
            bgrd = cfg->GetFloat('BGRD')
            PRINT, '  background at [', iFile, ',', iZ, '] = ', bgrd
        ENDIF
        IF bgrd GT self.bgrd THEN self.bgrd = bgrd
        self.nPhot += TOTAL((*data)[*,*,iZ]) - s*bgrd
        count += 1.0
    ENDFOR
ENDFOR
self.nPhot /= count
PRINT, '  self.nPhot=', self.nPhot
; create exactly one target array and put all position angles into
; that array (rescale all images to the same photometric intensity!)
self.obj = PTR_NEW(FLTARR(self.nx, self.ny))
self.objt = PTR_NEW(FLTARR(self.nx, self.ny))
self.objp = PTR_NEW(FLTARR(self.nx, self.ny))
nFiles = self.objFT->GetNFiles()
bgrd = 0.0
FOR iFile=0,nFiles-1 DO BEGIN
    file = self.objFT->GetFile(iFile)
    IF file->LoadData() EQ 0 THEN RETURN, 0
    data = file->GetData()
    FOR iZ=0, file->GetNZ()-1 DO BEGIN
        cfg = file->GetImageCFG(iZ)
        IF OBJ_VALID(cfg) THEN BEGIN
            bgrd = cfg->GetFloat('BGRD')
        ENDIF
        (*self.obj)[*,*] += ((*data)[*,*,iZ] - bgrd)*(1.0/(TOTAL((*data)[*,*,iZ], /DOUBLE) - s*bgrd))
    ENDFOR
ENDFOR
(*self.obj)[*,*] *= self.nPhot/TOTAL((*self.obj)[*,*], /DOUBLE) ; normalize it to integral nPhot
(*self.obj)[*,*] += self.bgrd
; create the corrections array
self.corr = PTR_NEW(FLTARR(self.nx, self.ny))
; create the target arrays
self.tar = PTR_NEW(FLTARR(self.nx, self.ny))
self.tar1 = PTR_NEW(FLTARR(self.nx, self.ny))
PRINT, 'ldrsDECONV_IBD::Prepare', self.nx, self.ny, self.nPA, self.nPhot, self.nPhot/s
(*self.rec)[*] = self.nPhot/s
(*self.tar)[*] = (*self.rec)[*]
LDRS_GUI_DISPATCH, 'ldrsIMAGE', 'ShowImage', IMAGE=self.rec, NAME='Reconstruction'
;WAIT, 10
RETURN, 1
END

FUNCTION ldrsDECONV_IBD::ExecuteStep
;PRINT, 'ldrsDECONV_IBD::ExecuteStep: self.iIter=', self.iIter
IF NOT self->ldrsDECONVOLVE::ExecuteStep() THEN RETURN, 0
currectExcept = !EXCEPT
!EXCEPT = 0
void = CHECK_MATH()
;self->IterateTar
;self->IteratePsf
; 1. convolve the current reconstruction (self.tar) with the current
; psf (self.psf) including number of photons and background
; tmpImg = target convolve psf + background
psfFft = FFT(*self.psf)
tarFft = FFT(*self.tar)
tmpImg = tarFft*psfFft
tmpImg = SHIFT(ABS(FFT(tmpImg, /INVERSE)), self.nx/2, self.ny/2)
tmpImg *= self.nPhot/TOTAL(tmpImg, /DOUBLE)
tmpImg += self.bgrd
; 2.a. calculate the correction for the target (normal RL)
corrTar = FFT((*self.obj)/tmpImg)*CONJ(psfFft)
corrTar = SHIFT(ABS(FFT(corrTar, /INVERSE)), self.nx/2, self.ny/2)
; 2.b. calculate the next target
(*self.tar1)[*] = (*self.tar)*corrTar
(*self.tar1) *= self.nPhot/TOTAL(*self.tar1, /DOUBLE)
; 3.a. calculate the correction for the PSF
corrPsf = FFT((*self.obj)/tmpImg)*CONJ(tarFft)
corrPsf = SHIFT(ABS(FFT(corrPsf, /INVERSE)), self.nx/2, self.ny/2)
; 3.b. calculate the next PSF
(*self.psf1)[*] = (*self.psf)*corrPsf
(*self.psf1) *= 1.0/TOTAL(*self.psf1, /DOUBLE)
; 4.a. calculate the next images for the target iteration
(*self.objt)[*] = SHIFT(ABS(FFT(FFT(*self.tar1)*FFT(*self.psf), /DOUBLE, /INVERSE)), self.nx/2, self.ny/2)
(*self.objt)[*] *= self.nPhot/TOTAL(*self.objt, /DOUBLE) 
(*self.objt)[*] += self.bgrd
; 4.b. calculate the next images for the PSF iteration
(*self.objp)[*] = SHIFT(ABS(FFT(FFT(*self.tar)*FFT(*self.psf1), /DOUBLE, /INVERSE)), self.nx/2, self.ny/2)
(*self.objp)[*] *= self.nPhot/TOTAL(*self.objp, /DOUBLE) 
(*self.objp)[*] += self.bgrd
; 5. calculate the error for both iterations
errTarget = SQRT(TOTAL((*self.objt - *self.obj)^2)/TOTAL(*self.obj^2))
errPsf = SQRT(TOTAL((*self.objp - *self.obj)^2)/TOTAL(*self.obj^2))
; 6. calculate the needed modification
modTarget = TOTAL(ABS(*self.tar1 - *self.tar), /DOUBLE)/TOTAL(*self.tar, /DOUBLE)
modPsf = TOTAL(ABS(*self.psf1 - *self.psf), /DOUBLE)/TOTAL(*self.psf, /DOUBLE)
IF errPsf LT errTarget THEN BEGIN
   PRINT, 'Iteration', self.iIter, ' psf => ', errPsf, '(', modPsf, ') target => ', errTarget, '(', modTarget, ') => improve psf'
   (*self.rec)[*] = (*self.tar)[*]
   (*self.psf)[*] = (*self.psf1)[*]
ENDIF ELSE BEGIN
   PRINT, 'Iteration', self.iIter, ' psf => ', errPsf, '(', modPsf, ') target => ', errTarget, '(', modTarget, ') => improve target'
   (*self.rec)[*] = (*self.tar1)[*]
   (*self.tar)[*] = (*self.tar1)[*]
ENDELSE
status = CHECK_MATH()
IF (status AND NOT 32) NE 0 THEN MESSAGE, 'IDL CHECK_MATH() error: ' + STRTRIM(status, 2)
!EXCEPT = currectExcept
;PRINT, 'ldrsDECONV_IBD::ExecuteStep, N_ELEMENTS((*self.rec))=', N_ELEMENTS((*self.rec))
;PRINT, 'ldrsDECONV_IBD::ExecuteStep, SIZE((*self.rec))=', SIZE((*self.rec))
RETURN, 1
END

PRO ldrsDECONV_IBD::Load, cfg
self->ldrsDECONVOLVE::Load, cfg
IF OBJ_VALID(cfg) THEN BEGIN
    self.psfIter = cfg->GetInt('PSFITER', DEFAULT=0)
    self.tarIter = cfg->GetInt('TARITER', DEFAULT=20)
ENDIF ELSE BEGIN
    self.psfIter = 0
    self.tarIter = 20
ENDELSE
END

PRO ldrsDECONV_IBD::Store, cfg
self->ldrsDECONVOLVE::Store, cfg
IF OBJ_VALID(cfg) THEN BEGIN
    cfg->SetInt, 'PSFITER', self.psfIter
    cfg->SetInt, 'TARITER', self.tarIter
ENDIF
END

PRO ldrsDECONV_IBD::ReadDynamicContent
self->ldrsDECONVOLVE::ReadDynamicContent
WIDGET_CONTROL, self.dtarIter, SET_VALUE=STRING(self.tarIter)
WIDGET_CONTROL, self.dpsfIter, SET_VALUE=STRING(self.psfIter)
END

PRO ldrsDECONV_IBD::WriteDynamicContent
self->ldrsDECONVOLVE::WriteDynamicContent
WIDGET_CONTROL, self.dtarIter, GET_VALUE=str
self.tarIter = FIX(str)
WIDGET_CONTROL, self.dpsfIter, GET_VALUE=str
self.psfIter = FIX(str)
END

PRO ldrsDECONV_IBD::CreateContent
self->ldrsDECONVOLVE::CreateContent
row = self->GetFreeRow()
lab             = self->CreateLabel(0, row, 'target iter:', /HALIGN_RIGHT)
self.dtarIter   = self->CreateEntry(2, row, VALUE='10')
row++
lab             = self->CreateLabel(0, row, 'psf iter:', /HALIGN_RIGHT)
self.dpsfIter   = self->CreateEntry(2, row, VALUE='5')
row++
END

PRO ldrsDECONV_IBD__define
struct = { ldrsDECONV_IBD, $
           INHERITS ldrsDECONVOLVE, $
           dpsfIter:0, dtarIter:0, $
           psfIter:0, tarIter:0, $
           psf:PTR_NEW(), $
           psf1:PTR_NEW(), $
           tar:PTR_NEW(), $
           tar1:PTR_NEW(), $
           obj:PTR_NEW(), $
           objp:PTR_NEW(), $
           objt:PTR_NEW(), $
           corr:PTR_NEW(), $
           bgrd:0.0 $
         }
END
